#
# Jason S. McMullan (c) 2019
#

import argparse
import os
import sys

import chitubox.session


def _human_value(value):
    if value < 2048:
        return str(value) + "b"
    elif value < 2048 * 1024:
        return "%.2f" % (value / 1024) + "Kb"
    elif value < 2048 * 1024 * 1024:
        return "%.2f" % (value / 1024 / 1024) + "Mb"
    else:
        return "%.2f" % (value / 1024 / 1024 / 1024) + "Gb"


def _progress(filename="", offset=0, total=1):
    sys.stdout.write("%s: %s/%s\r" %
                     (filename, _human_value(offset), _human_value(total)))
    if offset == total:
        sys.stdout.write("\n")


def cli(argv=[]):
    """Command line interface to ChuTuBox LCD Printers
    """

    parser = argparse.ArgumentParser()

    parser.add_argument("--ip", "-i", action="store", required=True)

    group = parser.add_mutually_exclusive_group()
    group.add_argument("--list", "-L", action="store_true", default=False)
    group.add_argument("--query", "-Q", action="store_true", default=False)
    group.add_argument("--axis-z", "-Z", action="store_true", default=False)
    group.add_argument("--version", "-V", action="store_true", default=False)
    group.add_argument("--print", "-P", action="store_true", default=False)
    group.add_argument("--upload", "-U", action="store_true", default=False)
    group.add_argument("--download", "-D", action="store_true", default=False)

    parser.add_argument("files", nargs="*")

    args = parser.parse_args(argv)

    session = chitubox.session.Session(ip=args.ip)
    session.progress = _progress

    if args.list:
        fileset = session.list("/", recurse=True)
        for fname, length in fileset:
            print(fname, _human_value(length))
    elif args.query:
        config = session.query_config()
        print("Resolution: x:%.4f, y:%.4f, z:%.5f mm" %
              (config['X'], config['Y'], config['Z']))
        try:
            status = session.print_status()
            print(status)
            status = session.query_status()
            print(status)
        except:
            print("Not printing")
    elif args.axis_z:
        print(session.query_axes()['Z'])
    elif args.version:
        print(session.query_version())
    elif args.print:
        print(session.start_print(args.print))
    elif args.upload:
        for filename in args.files:
            with open(filename, "rb") as fd:
                session.upload(os.path.basename(filename), fd)
    elif args.download:
        for filename in args.files:
            with open(os.path.basename(filename), "wb") as fd:
                session.download(filename, fd)
    else:
        parser.help()
        return 255

    return 0


if __name__ == "__main__":
    sys.exit(cli(sys.argv[1:]))



#
# Copyright 2019, Jason S. McMullan <jason.mcmullan@gmail.com>
#

import socket


class Udp(object):
    """Create a UDP socket connection to a ChiTuBox LCD printer (Anycubic Photon, EPAX X1, etc)
    """

    def __init__(self):
        self._socket = None
        self._send_ip = None
        self._send_port = 3000
        self._bind_port = 3001
        self._encoding = "utf-8"

    def __del__(self):
        self.disconnect()

    def connect(self, ip=None):
        """Connect to a remote printer
        :param ip: IP address
        """

        if not ip:
            self._send_ip = None
            self._socket = None
            self._mtu = 0
        else:
            self._send_ip = ip

            sock = socket.socket(family=socket.AF_INET,
                                 type=socket.SOCK_DGRAM, proto=0)
            sock.connect((self._send_ip, self._send_port))
            self._socket = sock
            self._mtu = 1500

        return True

    def disconnect(self):
        """Disconnect from a remote printer
        """

        if self._socket is not None:
            self._socket.close()
        self._socket = None
        self._send_ip = None
        self._mtu = 0

    def encoding(self, encode=None):
        current_encoding = self._encoding

        if encode:
            self._encoding = encode

        return current_encoding

    def recv(self):
        if not self._socket:
            raise RuntimeError("No remote session to printer")

        return self._socket.recv(self._mtu)

    def send(self, data=b''):
        if not self._socket:
            raise RuntimeError("No remote session to printer")

        return self._socket.send(data)

    def command(self, gcode=""):

        data = gcode.encode(self._encoding)

        self.send(data + b'\x00\x00')

    def response(self):
        resp = self._socket.recv(self._mtu)
        return resp.decode(self._encoding)



#
# Jason S. McMullan (c) 2019
#

import struct
import os

import chitubox.network


def _progress(filename="", offset=0, size=0):
    pass


class Session(object):
    """Start a ChiTuBox session
    """

    def __init__(self, ip=None):
        self._net = chitubox.network.Udp()
        self._net.connect(ip=ip)
        self.progress = _progress

        # Read the config
        config = self.query_config()

        # Update the network session's encoding to match the remote
        self._net.encoding(config['U'])

    def _parse_value(self, val=""):
        if len(val) == 0:
            return None
        elif val.startswith("'") and val.endswith("'"):
            return val.strip("'")
        elif '/' in val:
            vals = []
            for subval in val.split("/"):
                vals.append(self._parse_value(subval))
            return vals
        elif '.' in val:
            return float(val)
        else:
            return int(val)

    def _parse_fields(self, s=""):
        fields = {}
        for field in s.split(" "):
            attr, val = field.split(":", 2)
            fields[attr] = self._parse_value(val)
        return fields

    def response(self, fields=True, command=""):
        result = []
        fieldset = {}
        while True:
            resp = self._net.response().strip()
            if resp.startswith("Error"):
                raise RuntimeError(command + ": " + resp)
            if resp.startswith("resend "):
                offset, rest = resp.split(",")
                dummy, offset = offset.split(" ")
                fieldset["offset"] = int(offset)
                dummy, error = rest.split(":")
                fieldset["offset error"] = int(error)
                return ["resend"], fieldset
            if resp == "ok":
                break
            if resp.startswith("ok "):
                if not fields:
                    result.append(resp)
                else:
                    fieldset = self._parse_fields(resp[3:])
                break
            else:
                result.append(resp)
        return result, fieldset

    def gcode(self, gcode="", fields=True):
        self._net.command(gcode=gcode)
        return self.response(fields=fields, command=gcode)

    def list(self, root="/", recurse=False):
        result, fields = self.gcode("M20 '" + root + "'")
        if len(result) < 2:
            raise RuntimeError("M20: Response too short")
        elif result[0] != 'Begin file list':
            raise RuntimeError("M20: Unexpected header")
        elif result[-1] != 'End file list':
            raise RuntimeError("M20: Unexpected footer")

        flist = []

        for x in result[1:-1]:
            if x.startswith("->"):
                flist += self.list(os.path.join(root, x[2:]))
            else:
                f, s = x.rsplit(" ", 2)
                s = int(s)
                flist.append((os.path.join(root, f), s))

        return flist

    def query_version(self):
        result, _ = self.gcode("M4002", fields=False)

        return result[-1].split(" ", 2)[1]

    def query_config(self):
        result, fields = self.gcode("M4001")

        return fields

    def query_status(self):
        result, fields = self.gcode("M4000")
        return fields

    def print_status(self):
        result, fields = self.gcode("M27")

        return fields

    def query_axes(self):
        result, fields = self.gcode("M114")
        return fields

    def start_print(self, filename=None):
        result, fields = self.gcode("M6030 '" + filename + "'")
        return fields

    def download(self, filename=None, fd=None):
        # Abort any prior download
        self.gcode("M22")
        result, fields = self.gcode("M6032 '" + filename + "'")
        length = fields['L']

        # Download data
        curr = 0
        while curr < length:
            self._net.command("M3000 I%d" % (curr))
            block = self._net.recv()
            offset, csum, verify = struct.unpack("<LBB", block[-6:])

            check = 0
            for x in block[:-2]:
                check ^= x

            if check != csum:
                raise RuntimeError(
                    "Checksum failed (%02x != %02x) at offset %d" % (check, csum, curr))

            if verify != 0x83:
                raise RuntimeError("Verification failed at offset %d" % (curr))

            if curr != offset:
                continue

            block = block[:-6]
            curr += len(block)
            fd.write(block)

            self.progress(filename, curr, length)

        self.gcode("M22")

    def upload(self, filename=None, fd=None):
        # Abort any prior download
        self.gcode("M22")
        result, fields = self.gcode("M28 " + filename)

        length = fd.seek(0, 2)
        fd.seek(0)
        while True:
            offset = fd.tell()
            block = fd.read(0x500)
            if len(block) == 0:
                break

            block += struct.pack("<L", offset)

            check = 0
            for x in block:
                check ^= x

            block += struct.pack("<BB", check, 0x83)
            self._net.send(block)
            result, fields = self.response(
                fields=False, command="Upload @%d" % (offset))
            if len(result) > 0 and result[0] == "resend":
                fd.seek(fields["offset"])

            self.progress(filename, offset, length)

        self.progress(filename, length, length)

        result, fields = self.gcode("M29")



# -*- coding: utf-8 -*-

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="chitubox",
    version="0.1.2",
    author="Jason S. McMullan",
    author_email="jason.mcmullan@gmail.com",
    description="Utility to manage networked ChiTuBox based LCD Resin printers",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ezrec/python-chitubox",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.5',
    zip_safe=True,
    include_package_data=False,
    entry_points={
        "console_scripts": [
            "chitucli=chitubox:cli"
        ],
    }
)
